# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 University of New Hampshire

"""DPDK checksum offload testing suite.

This suite verifies L3/L4 checksum offload features of the Poll Mode Driver.
On the Rx side, IPv4 and UDP/TCP checksum by hardware is checked to ensure
checksum flags match expected flags. On the Tx side, IPv4/UDP, IPv4/TCP,
IPv6/UDP, and IPv6/TCP insertion by hardware is checked to checksum flags
match expected flags.

"""

from typing import List

from scapy.layers.inet import IP, TCP, UDP  # type: ignore[import-untyped]
from scapy.layers.inet6 import IPv6  # type: ignore[import-untyped]
from scapy.layers.l2 import Dot1Q, Ether  # type: ignore[import-untyped]
from scapy.layers.sctp import SCTP  # type: ignore[import-untyped]
from scapy.packet import Packet, Raw  # type: ignore[import-untyped]

from framework.params.testpmd import SimpleForwardingModes
from framework.remote_session.testpmd_shell import (
    ChecksumOffloadOptions,
    PacketOffloadFlag,
    TestPmdShell,
)
from framework.test_suite import TestSuite, func_test
from framework.testbed_model.capability import NicCapability, TopologyType, requires


@requires(topology_type=TopologyType.two_links)
@requires(NicCapability.RX_OFFLOAD_CHECKSUM)
@requires(NicCapability.RX_OFFLOAD_IPV4_CKSUM)
@requires(NicCapability.RX_OFFLOAD_UDP_CKSUM)
@requires(NicCapability.RX_OFFLOAD_TCP_CKSUM)
class TestChecksumOffload(TestSuite):
    """Checksum offload test suite.

    This suite consists of 6 test cases:
    1. Insert checksum on transmit packet
    2. Do not insert checksum on transmit packet
    3. Hardware checksum check L4 Rx
    4. Hardware checksum check L3 Rx
    5. Validate Rx checksum valid flags
    6. Checksum offload with VLAN
    7. Checksum offload with SCTP

    """

    def send_packets_and_verify(
        self, packet_list: List[Packet], load: str, should_receive: bool
    ) -> None:
        """Iterates through a list of packets and verifies they are received.

        Args:
            packet_list: List of Scapy packets to send and verify.
            load: Raw layer load attribute in the sent packet.
            should_receive: Indicates whether the packet should be received
                by the traffic generator.
        """
        for i in range(0, len(packet_list)):
            received_packets = self.send_packet_and_capture(packet=packet_list[i])
            received = any(
                packet.haslayer(Raw) and load in str(packet.load) for packet in received_packets
            )
            self.verify(
                received == should_receive,
                f"Packet was {'dropped' if should_receive else 'received'}",
            )

    def send_packet_and_verify_checksum(
        self, packet: Packet, goodL4: bool, goodIP: bool, testpmd: TestPmdShell, id: str
    ) -> None:
        """Send packet and verify verbose output matches expected output.

        Args:
            packet: Scapy packet to send to DUT.
            goodL4: Verifies RTE_MBUF_F_RX_L4_CKSUM_GOOD in verbose output
                if :data:`True`, or RTE_MBUF_F_RX_L4_CKSUM_UNKNOWN if :data:`False`.
            goodIP: Verifies RTE_MBUF_F_RX_IP_CKSUM_GOOD in verbose output
                if :data:`True`, or RTE_MBUF_F_RX_IP_CKSUM_UNKNOWN if :data:`False`.
            testpmd: Testpmd shell session to analyze verbose output of.
            id: The destination mac address that matches the sent packet in verbose output.
        """
        testpmd.start()
        self.send_packet_and_capture(packet=packet)
        verbose_output = testpmd.extract_verbose_output(testpmd.stop())
        for packet in verbose_output:
            if packet.dst_mac == id:
                isIP = PacketOffloadFlag.RTE_MBUF_F_RX_IP_CKSUM_GOOD in packet.ol_flags
                isL4 = PacketOffloadFlag.RTE_MBUF_F_RX_L4_CKSUM_GOOD in packet.ol_flags
        self.verify(isL4 == goodL4, "Layer 4 checksum flag did not match expected checksum flag.")
        self.verify(isIP == goodIP, "IP checksum flag did not match expected checksum flag.")

    def setup_hw_offload(self, testpmd: TestPmdShell) -> None:
        """Sets IP, UDP, and TCP layers to hardware offload.

        Args:
            testpmd: Testpmd shell to configure.
        """
        testpmd.stop_all_ports()
        offloads = (
            ChecksumOffloadOptions.ip | ChecksumOffloadOptions.udp | ChecksumOffloadOptions.tcp
        )
        testpmd.csum_set_hw(layers=offloads, port_id=0)
        testpmd.start_all_ports()

    @func_test
    def test_insert_checksums(self) -> None:
        """Enable checksum offload insertion and verify packet reception."""
        mac_id = "00:00:00:00:00:01"
        payload = "xxxxx"
        packet_list = [
            Ether(dst=mac_id) / IP() / UDP() / Raw(payload),
            Ether(dst=mac_id) / IP() / TCP() / Raw(payload),
            Ether(dst=mac_id) / IPv6(src="::1") / UDP() / Raw(payload),
            Ether(dst=mac_id) / IPv6(src="::1") / TCP() / Raw(payload),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            self.setup_hw_offload(testpmd=testpmd)
            testpmd.start()
            self.send_packets_and_verify(packet_list=packet_list, load=payload, should_receive=True)
            for i in range(0, len(packet_list)):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
                )

    @func_test
    def test_no_insert_checksums(self) -> None:
        """Disable checksum offload insertion and verify packet reception."""
        mac_id = "00:00:00:00:00:01"
        payload = "xxxxx"
        packet_list = [
            Ether(dst=mac_id) / IP() / UDP() / Raw(payload),
            Ether(dst=mac_id) / IP() / TCP() / Raw(payload),
            Ether(dst=mac_id) / IPv6(src="::1") / UDP() / Raw(payload),
            Ether(dst=mac_id) / IPv6(src="::1") / TCP() / Raw(payload),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            testpmd.start()
            self.send_packets_and_verify(packet_list=packet_list, load=payload, should_receive=True)
            for i in range(0, len(packet_list)):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
                )

    @func_test
    def test_l4_rx_checksum(self) -> None:
        """Tests L4 Rx checksum in a variety of scenarios."""
        mac_id = "00:00:00:00:00:01"
        packet_list = [
            Ether(dst=mac_id) / IP() / UDP(),
            Ether(dst=mac_id) / IP() / TCP(),
            Ether(dst=mac_id) / IP() / UDP(chksum=0xF),
            Ether(dst=mac_id) / IP() / TCP(chksum=0xF),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            self.setup_hw_offload(testpmd=testpmd)
            for i in range(0, 2):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
                )
            for i in range(2, 4):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=False, goodIP=True, testpmd=testpmd, id=mac_id
                )

    @func_test
    def test_l3_rx_checksum(self) -> None:
        """Tests L3 Rx checksum hardware offload."""
        mac_id = "00:00:00:00:00:01"
        packet_list = [
            Ether(dst=mac_id) / IP() / UDP(),
            Ether(dst=mac_id) / IP() / TCP(),
            Ether(dst=mac_id) / IP(chksum=0xF) / UDP(),
            Ether(dst=mac_id) / IP(chksum=0xF) / TCP(),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            self.setup_hw_offload(testpmd=testpmd)
            for i in range(0, 2):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
                )
            for i in range(2, 4):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=False, testpmd=testpmd, id=mac_id
                )

    @func_test
    def test_validate_rx_checksum(self) -> None:
        """Verify verbose output of Rx packets matches expected behavior."""
        mac_id = "00:00:00:00:00:01"
        packet_list = [
            Ether(dst=mac_id) / IP() / UDP(),
            Ether(dst=mac_id) / IP() / TCP(),
            Ether(dst=mac_id) / IPv6(src="::1") / UDP(),
            Ether(dst=mac_id) / IPv6(src="::1") / TCP(),
            Ether(dst=mac_id) / IP(chksum=0x0) / UDP(chksum=0xF),
            Ether(dst=mac_id) / IP(chksum=0x0) / TCP(chksum=0xF),
            Ether(dst=mac_id) / IPv6(src="::1") / UDP(chksum=0xF),
            Ether(dst=mac_id) / IPv6(src="::1") / TCP(chksum=0xF),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            self.setup_hw_offload(testpmd=testpmd)
            for i in range(0, 4):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
                )
            for i in range(4, 6):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=False, goodIP=False, testpmd=testpmd, id=mac_id
                )
            for i in range(6, 8):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=False, goodIP=True, testpmd=testpmd, id=mac_id
                )

    @requires(NicCapability.RX_OFFLOAD_VLAN)
    @func_test
    def test_vlan_checksum(self) -> None:
        """Test VLAN Rx checksum hardware offload and verify packet reception."""
        mac_id = "00:00:00:00:00:01"
        payload = "xxxxx"
        packet_list = [
            Ether(dst=mac_id) / Dot1Q(vlan=1) / IP(chksum=0x0) / UDP(chksum=0xF) / Raw(payload),
            Ether(dst=mac_id) / Dot1Q(vlan=1) / IP(chksum=0x0) / TCP(chksum=0xF) / Raw(payload),
            Ether(dst=mac_id) / Dot1Q(vlan=1) / IPv6(src="::1") / UDP(chksum=0xF) / Raw(payload),
            Ether(dst=mac_id) / Dot1Q(vlan=1) / IPv6(src="::1") / TCP(chksum=0xF) / Raw(payload),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            self.setup_hw_offload(testpmd=testpmd)
            testpmd.start()
            self.send_packets_and_verify(packet_list=packet_list, load=payload, should_receive=True)
            for i in range(0, 2):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=False, goodIP=False, testpmd=testpmd, id=mac_id
                )
            for i in range(2, 4):
                self.send_packet_and_verify_checksum(
                    packet=packet_list[i], goodL4=False, goodIP=True, testpmd=testpmd, id=mac_id
                )

    @requires(NicCapability.RX_OFFLOAD_SCTP_CKSUM)
    @func_test
    def test_validate_sctp_checksum(self) -> None:
        """Test SCTP Rx checksum hardware offload and verify packet reception."""
        mac_id = "00:00:00:00:00:01"
        packet_list = [
            Ether(dst=mac_id) / IP() / SCTP(),
            Ether(dst=mac_id) / IP() / SCTP(chksum=0xF),
        ]
        with TestPmdShell(node=self.sut_node, enable_rx_cksum=True) as testpmd:
            testpmd.set_forward_mode(SimpleForwardingModes.csum)
            testpmd.set_verbose(level=1)
            testpmd.csum_set_hw(layers=ChecksumOffloadOptions.sctp)
            testpmd.start()
            self.send_packet_and_verify_checksum(
                packet=packet_list[0], goodL4=True, goodIP=True, testpmd=testpmd, id=mac_id
            )
            self.send_packet_and_verify_checksum(
                packet=packet_list[1], goodL4=False, goodIP=True, testpmd=testpmd, id=mac_id
            )



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2020 Intel Corporation

"""
Configure an entire Intel DSA instance, using idxd kernel driver, for DPDK use
"""

import sys
import argparse
import os
import os.path


class SysfsDir:
    verbose = False

    "Used to read/write paths in a sysfs directory"
    def __init__(self, path):
        self.path = path

    def read_int(self, filename):
        "Return a value from sysfs file"
        if SysfsDir.verbose:
            print(f"Reading '{filename}' in {self.path}")
        with open(os.path.join(self.path, filename)) as f:
            return int(f.readline())

    def write_values(self, values):
        "write dictionary, where key is filename and value is value to write"
        for filename, contents in values.items():
            if SysfsDir.verbose:
                print(f"Writing '{contents}' to '{filename}' in {self.path}")
            with open(os.path.join(self.path, filename), "w") as f:
                f.write(str(contents))


def get_drv_dir(dtype):
    "Get the sysfs path for the driver, either 'idxd' or 'user'"
    drv_dir = "/sys/bus/dsa/drivers/" + dtype
    if not os.path.exists(drv_dir):
        return "/sys/bus/dsa/drivers/dsa"
    return drv_dir


def reset_device(dsa_id):
    "Reset the DSA device and all its queues"
    drv_dir = SysfsDir(get_drv_dir("idxd"))
    drv_dir.write_values({"unbind": f"dsa{dsa_id}"})


def get_pci_dir(pci):
    "Search for the sysfs directory of the PCI device"
    base_dir = '/sys/bus/pci/devices/'
    for path, dirs, files in os.walk(base_dir):
        for dir in dirs:
            if pci in dir:
                return os.path.join(base_dir, dir)
    sys.exit(f"Could not find sysfs directory for device {pci}")


def get_dsa_id(pci):
    "Get the DSA instance ID using the PCI address of the device"
    pci_dir = get_pci_dir(pci)
    for path, dirs, files in os.walk(pci_dir):
        for dir in dirs:
            if dir.startswith('dsa') and 'wq' not in dir:
                return int(dir[3:])
    sys.exit(f"Could not get device ID for device {pci}")


def parse_wq_opts(wq_opts):
    "Parse user-specified queue configuration, creating a dict of options"
    try:
        return {o.split('=')[0]: o.split('=')[1] for o in wq_opts}
    except ValueError:
        sys.exit("Invalid --wq-option format, use format 'option=value'")


def configure_dsa(dsa_id, args):
    "Configure the DSA instance with appropriate number of queues"
    dsa_dir = SysfsDir(f"/sys/bus/dsa/devices/dsa{dsa_id}")

    max_groups = dsa_dir.read_int("max_groups")
    max_engines = dsa_dir.read_int("max_engines")
    max_queues = dsa_dir.read_int("max_work_queues")
    max_work_queues_size = dsa_dir.read_int("max_work_queues_size")

    nb_queues = min(args.q, max_queues)
    if args.q > nb_queues:
        print(f"Setting number of queues to max supported value: {max_queues}")

    # we want one engine per group, and no more engines than queues
    nb_groups = min(max_engines, max_groups, nb_queues)
    for grp in range(nb_groups):
        dsa_dir.write_values({f"engine{dsa_id}.{grp}/group_id": grp})

    # configure each queue
    for q in range(nb_queues):
        wqcfg = {"group_id": q % nb_groups,
                 "type": "user",
                 "mode": "dedicated",
                 "name": f"{args.prefix}_wq{dsa_id}.{q}",
                 "priority": 1,
                 "max_batch_size": 1024,
                 "size": int(max_work_queues_size / nb_queues)}
        wq_dir = SysfsDir(os.path.join(dsa_dir.path, f"wq{dsa_id}.{q}"))
        if os.path.exists(os.path.join(wq_dir.path, f"driver_name")):
            wqcfg.update({"driver_name": "user"})
        wqcfg.update(parse_wq_opts(args.wq_option))
        wq_dir.write_values(wqcfg)

    # enable device and then queues
    idxd_dir = SysfsDir(get_drv_dir("idxd"))
    idxd_dir.write_values({"bind": f"dsa{dsa_id}"})

    user_dir = SysfsDir(get_drv_dir("user"))
    for q in range(nb_queues):
        user_dir.write_values({"bind": f"wq{dsa_id}.{q}"})


def main(args):
    "Main function, does arg parsing and calls config function"
    arg_p = argparse.ArgumentParser(
        description="Configure whole DSA device instance for DPDK use")
    arg_p.add_argument('dsa_id',
                       help="Specify DSA instance either via DSA instance number or PCI address")
    arg_p.add_argument('-q', metavar='queues', type=int, default=255,
                       help="Number of queues to set up")
    arg_p.add_argument('--name-prefix', metavar='prefix', dest='prefix',
                       default="dpdk",
                       help="Prefix for workqueue name to mark for DPDK use [default: 'dpdk']")
    arg_p.add_argument('--wq-option', action='append', default=[],
                       help="Provide additional config option for queues (format 'x=y')")
    arg_p.add_argument('--verbose', '-v', action='store_true',
                       help="Provide addition info on tasks being performed")
    arg_p.add_argument('--reset', action='store_true',
                       help="Reset DSA device and its queues")
    parsed_args = arg_p.parse_args(args[1:])

    dsa_id = parsed_args.dsa_id
    dsa_id = get_dsa_id(dsa_id) if ':' in dsa_id else dsa_id

    SysfsDir.verbose = parsed_args.verbose
    if parsed_args.reset:
        reset_device(dsa_id)
    else:
        configure_dsa(dsa_id, parsed_args)


if __name__ == "__main__":
    main(sys.argv)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023-2024 University of New Hampshire

"""Multi-segment packet scattering testing suite.

This testing suite tests the support of transmitting and receiving scattered packets.
This is shown by the Poll Mode Driver being able to forward
scattered multi-segment packets composed of multiple non-contiguous memory buffers.
To ensure the receipt of scattered packets,
the DMA rings of the port's Rx queues must be configured
with mbuf data buffers whose size is less than the maximum length.

If it is the case that the Poll Mode Driver can forward scattered packets which it receives,
then this suffices to show the Poll Mode Driver is capable of both receiving and transmitting
scattered packets.
"""

import struct

from scapy.layers.inet import IP  # type: ignore[import-untyped]
from scapy.layers.l2 import Ether  # type: ignore[import-untyped]
from scapy.packet import Raw  # type: ignore[import-untyped]
from scapy.utils import hexstr  # type: ignore[import-untyped]

from framework.params.testpmd import SimpleForwardingModes
from framework.remote_session.testpmd_shell import TestPmdShell
from framework.test_suite import TestSuite, func_test
from framework.testbed_model.capability import NicCapability, requires


@requires(NicCapability.RX_OFFLOAD_SCATTER)
class TestPmdBufferScatter(TestSuite):
    """DPDK PMD packet scattering test suite.

    Configure the Rx queues to have mbuf data buffers
    whose sizes are smaller than the maximum packet size.
    Specifically, set mbuf data buffers to have a size of 2048
    to fit a full 1512-byte (CRC included) ethernet frame in a mono-segment packet.
    The testing of scattered packets is done by sending a packet
    whose length is greater than the size of the configured size of mbuf data buffers.
    There are a total of 5 packets sent within test cases
    which have lengths less than, equal to, and greater than the mbuf size.
    There are multiple packets sent with lengths greater than the mbuf size
    in order to test cases such as:

    1. A single byte of the CRC being in a second buffer
       while the remaining 3 bytes are stored in the first buffer alongside packet data.
    2. The entire CRC being stored in a second buffer
       while all of the packet data is stored in the first.
    3. Most of the packet data being stored in the first buffer
       and a single byte of packet data stored in a second buffer alongside the CRC.
    """

    def set_up_suite(self) -> None:
        """Set up the test suite.

        Setup:
            Increase the MTU of both ports on the traffic generator to 9000
            to support larger packet sizes.
        """
        self.tg_node.main_session.configure_port_mtu(9000, self._tg_port_egress)
        self.tg_node.main_session.configure_port_mtu(9000, self._tg_port_ingress)

    def scatter_pktgen_send_packet(self, pktsize: int) -> str:
        """Generate and send a packet to the SUT then capture what is forwarded back.

        Generate an IP packet of a specific length and send it to the SUT,
        then capture the resulting received packet and extract its payload.
        The desired length of the packet is met by packing its payload
        with the letter "X" in hexadecimal.

        Args:
            pktsize: Size of the packet to generate and send.

        Returns:
            The payload of the received packet as a string.
        """
        packet = Ether() / IP() / Raw()
        packet.getlayer(2).load = ""
        payload_len = pktsize - len(packet) - 4
        payload = ["58"] * payload_len
        # pack the payload
        for X_in_hex in payload:
            packet.load += struct.pack("=B", int("%s%s" % (X_in_hex[0], X_in_hex[1]), 16))
        received_packets = self.send_packet_and_capture(packet)
        self.verify(len(received_packets) > 0, "Did not receive any packets.")
        load = hexstr(received_packets[0].getlayer(2), onlyhex=1)

        return load

    def pmd_scatter(self, mbsize: int) -> None:
        """Testpmd support of receiving and sending scattered multi-segment packets.

        Support for scattered packets is shown by sending 5 packets of differing length
        where the length of the packet is calculated by taking mbuf-size + an offset.
        The offsets used in the test are -1, 0, 1, 4, 5 respectively.

        Test:
            Start testpmd and run functional test with preset mbsize.
        """
        with TestPmdShell(
            self.sut_node,
            forward_mode=SimpleForwardingModes.mac,
            mbcache=200,
            mbuf_size=[mbsize],
            max_pkt_len=9000,
            tx_offloads=0x00008000,
        ) as testpmd:
            testpmd.start()

            for offset in [-1, 0, 1, 4, 5]:
                recv_payload = self.scatter_pktgen_send_packet(mbsize + offset)
                self._logger.debug(
                    f"Payload of scattered packet after forwarding: \n{recv_payload}"
                )
                self.verify(
                    ("58 " * 8).strip() in recv_payload,
                    "Payload of scattered packet did not match expected payload with offset "
                    f"{offset}.",
                )

    @requires(NicCapability.SCATTERED_RX_ENABLED)
    @func_test
    def test_scatter_mbuf_2048(self) -> None:
        """Run the :meth:`pmd_scatter` test with `mbsize` set to 2048."""
        self.pmd_scatter(mbsize=2048)

    def tear_down_suite(self) -> None:
        """Tear down the test suite.

        Teardown:
            Set the MTU of the tg_node back to a more standard size of 1500.
        """
        self.tg_node.main_session.configure_port_mtu(1500, self._tg_port_egress)
        self.tg_node.main_session.configure_port_mtu(1500, self._tg_port_ingress)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 Arm Limited

"""Module representing the DPDK EAL-related parameters."""

from dataclasses import dataclass, field
from typing import Literal

from framework.params import Params, Switch
from framework.testbed_model.cpu import LogicalCoreList
from framework.testbed_model.port import Port
from framework.testbed_model.virtual_device import VirtualDevice


def _port_to_pci(port: Port) -> str:
    return port.pci


@dataclass(kw_only=True)
class EalParams(Params):
    """The environment abstraction layer parameters.

    Attributes:
        lcore_list: The list of logical cores to use.
        memory_channels: The number of memory channels to use.
        prefix: Set the file prefix string with which to start DPDK, e.g.: ``prefix="vf"``.
        no_pci: Switch to disable PCI bus, e.g.: ``no_pci=True``.
        vdevs: Virtual devices, e.g.::

            vdevs=[
                VirtualDevice('net_ring0'),
                VirtualDevice('net_ring1')
            ]

        ports: The list of ports to allow.
        other_eal_param: user defined DPDK EAL parameters, e.g.::

            ``other_eal_param='--single-file-segments'``
    """

    lcore_list: LogicalCoreList | None = field(default=None, metadata=Params.short("l"))
    memory_channels: int | None = field(default=None, metadata=Params.short("n"))
    prefix: str = field(default="dpdk", metadata=Params.long("file-prefix"))
    no_pci: Switch = None
    vdevs: list[VirtualDevice] | None = field(
        default=None, metadata=Params.multiple() | Params.long("vdev")
    )
    allowed_ports: list[Port] | None = field(
        default=None,
        metadata=Params.convert_value(_port_to_pci) | Params.multiple() | Params.short("a"),
    )
    blocked_ports: list[Port] | None = field(
        default=None,
        metadata=Params.convert_value(_port_to_pci) | Params.multiple() | Params.short("b"),
    )
    other_eal_param: Params | None = None
    _separator: Literal[True] = field(default=True, init=False, metadata=Params.short("-"))



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2015 Intel Corporation

from docutils import nodes
from packaging.version import Version
from sphinx import __version__ as sphinx_version
from os import listdir
from os import environ
from os.path import basename
from os.path import dirname
from os.path import join as path_join
from sys import argv, stderr, path

import configparser

try:
    import sphinx_rtd_theme
    html_theme = "sphinx_rtd_theme"
except ImportError:
    print('Install the sphinx ReadTheDocs theme for improved html documentation '
          'layout: https://sphinx-rtd-theme.readthedocs.io/',
          file=stderr)
    html_theme = "default"

stop_on_error = ('-W' in argv)

project = 'Data Plane Development Kit'
html_logo = '../logo/DPDK_logo_vertical_rev_small.png'
if Version(sphinx_version) >= Version('3.5'):
    html_permalinks = False
else:
    html_add_permalinks = ""
html_show_copyright = False
highlight_language = 'none'

release = environ.setdefault('DPDK_VERSION', "None")
version = release

master_doc = 'index'

# Maximum feature description string length
feature_str_len = 30

# Figures, tables and code-blocks automatically numbered if they have caption
numfig = True

# Configuration for man pages
man_pages = [("testpmd_app_ug/run_app", "testpmd",
              "tests for dpdk pmds", "", 1),
             ("tools/pdump", "dpdk-pdump",
              "enable packet capture on dpdk ports", "", 1),
             ("tools/proc_info", "dpdk-procinfo",
              "access dpdk port stats and memory info", "", 1),
             ("tools/pmdinfo", "dpdk-pmdinfo",
              "dump a PMDs hardware support info", "", 1),
             ("tools/devbind", "dpdk-devbind",
              "check device status and bind/unbind them from drivers", "", 8)]

# DTS API docs additional configuration
if environ.get('DTS_DOC_BUILD'):
    extensions = ['sphinx.ext.napoleon', 'sphinx.ext.autodoc']

    # Pydantic models require autodoc_pydantic for the right formatting
    try:
        import sphinxcontrib.autodoc_pydantic
        extensions.append("sphinxcontrib.autodoc_pydantic")
    except ImportError:
        print(
            "The DTS API doc dependencies are missing. "
            "The generated output won't be as intended, "
            "and autodoc may throw unexpected warnings.",
            file=stderr,
        )

    # Napoleon enables the Google format of Python doscstrings.
    napoleon_numpy_docstring = False
    napoleon_attr_annotations = True
    napoleon_preprocess_types = True

    # Autodoc pulls documentation from code.
    autodoc_default_options = {
        'members': True,
        'member-order': 'bysource',
        'show-inheritance': True,
    }
    autodoc_class_signature = 'separated'
    autodoc_typehints = 'both'
    autodoc_typehints_format = 'short'
    autodoc_typehints_description_target = 'documented'

    # DTS docstring options.
    add_module_names = False
    toc_object_entries = True
    toc_object_entries_show_parents = 'hide'
    # DTS Sidebar config.
    if html_theme == "sphinx_rtd_theme":
        html_theme_options = {
            'collapse_navigation': False,
            'navigation_depth': -1,  # unlimited depth
        }

    # Add path to DTS sources so that Sphinx can find them.
    dpdk_root = dirname(dirname(dirname(__file__)))
    path.append(path_join(dpdk_root, 'dts'))

    # Get missing DTS dependencies. Add path to buildtools to find the get_missing_imports function.
    path.append(path_join(dpdk_root, 'buildtools'))
    import importlib
    # Ignore missing imports from DTS dependencies, allowing us to build the docs without them.
    # There's almost no difference between docs built with and without dependencies.
    # The qualified names of imported objects are fully expanded with dependencies, such as:
    # fabric.Connection (without) vs. fabric.connection.Connection (with)
    autodoc_mock_imports = importlib.import_module('check-dts-requirements').get_missing_imports()


# ####### :numref: fallback ########
# The following hook functions add some simple handling for the :numref:
# directive for Sphinx versions prior to 1.3.1. The functions replace the
# :numref: reference with a link to the target (for all Sphinx doc types).
# It doesn't try to label figures/tables.
def numref_role(reftype, rawtext, text, lineno, inliner):
    """
    Add a Sphinx role to handle numref references. Note, we can't convert
    the link here because the doctree isn't build and the target information
    isn't available.
    """
    # Add an identifier to distinguish numref from other references.
    newnode = nodes.reference('',
                              '',
                              refuri='_local_numref_#%s' % text,
                              internal=True)
    return [newnode], []


def process_numref(app, doctree, from_docname):
    """
    Process the numref nodes once the doctree has been built and prior to
    writing the files. The processing involves replacing the numref with a
    link plus text to indicate if it is a Figure or Table link.
    """

    # Iterate over the reference nodes in the doctree.
    for node in doctree.traverse(nodes.reference):
        target = node.get('refuri', '')

        # Look for numref nodes.
        if target.startswith('_local_numref_#'):
            target = target.replace('_local_numref_#', '')

            # Get the target label and link information from the Sphinx env.
            data = app.builder.env.domains['std'].data
            docname, label, _ = data['labels'].get(target, ('', '', ''))
            relative_url = app.builder.get_relative_uri(from_docname, docname)

            # Add a text label to the link.
            if target.startswith('figure'):
                caption = 'Figure'
            elif target.startswith('table'):
                caption = 'Table'
            else:
                caption = 'Link'

            # New reference node with the updated link information.
            newnode = nodes.reference('',
                                      caption,
                                      refuri='%s#%s' % (relative_url, label),
                                      internal=True)
            node.replace_self(newnode)


def generate_overview_table(output_filename, table_id, section, table_name, title):
    """
    Function to generate the Overview Table from the ini files that define
    the features for each driver.

    The default features for the table and their order is defined by the
    'default.ini' file.

    """
    # Default warning string.
    warning = 'Warning generate_overview_table()'

    # Get the default features and order from the 'default.ini' file.
    ini_path = path_join(dirname(output_filename), 'features')
    config = configparser.ConfigParser()
    config.optionxform = str
    config.read(path_join(ini_path, 'default.ini'))
    default_features = config.items(section)

    # Create a dict of the valid features to validate the other ini files.
    valid_features = {}
    max_feature_length = 0
    for feature in default_features:
        key = feature[0]
        valid_features[key] = ' '
        max_feature_length = max(max_feature_length, len(key))

    # Get a list of driver ini files, excluding 'default.ini'.
    ini_files = [basename(file) for file in listdir(ini_path)
                 if file.endswith('.ini') and file != 'default.ini']
    ini_files.sort()

    # Build up a list of the table header names from the ini filenames.
    pmd_names = []
    for ini_filename in ini_files:
        name = ini_filename[:-4]
        name = name.replace('_vf', 'vf')
        pmd_names.append(name)
    if not pmd_names:
        # Add an empty column if table is empty (required by RST syntax)
        pmd_names.append(' ')

    # Pad the table header names.
    max_header_len = len(max(pmd_names, key=len))
    header_names = []
    for name in pmd_names:
        if '_vec' in name:
            pmd, vec = name.split('_')
            name = '{0:{fill}{align}{width}}vec'.format(pmd,
                    fill='.', align='<', width=max_header_len-3)
        else:
            name = '{0:{fill}{align}{width}}'.format(name,
                    fill=' ', align='<', width=max_header_len)
        header_names.append(name)

    # Create a dict of the defined features for each driver from the ini files.
    ini_data = {}
    for ini_filename in ini_files:
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(path_join(ini_path, ini_filename))

        # Initialize the dict with the default.ini value.
        ini_data[ini_filename] = valid_features.copy()

        # Check for a section.
        if not config.has_section(section):
            continue

        # Check for valid features names.
        for name, value in config.items(section):
            if name not in valid_features:
                print("{}: Unknown feature '{}' in '{}'".format(warning,
                                                                name,
                                                                ini_filename),
                                                                file=stderr)
                if stop_on_error:
                    raise Exception('Warning is treated as a failure')
                continue

            if value:
                # Get the first letter only.
                ini_data[ini_filename][name] = value[0]

    # Print out the RST Driver Overview table from the ini file data.
    outfile = open(output_filename, 'w')
    num_cols = len(header_names)

    print_table_css(outfile, table_id)
    print('.. _' + table_name + ':', file=outfile)
    print('.. table:: ' + table_name + '\n', file=outfile)
    print_table_header(outfile, num_cols, header_names, title)
    print_table_body(outfile, num_cols, ini_files, ini_data, default_features)


def print_table_header(outfile, num_cols, header_names, title):
    """ Print the RST table header. The header names are vertical. """
    print_table_divider(outfile, num_cols)

    line = ''
    for name in header_names:
        line += ' ' + name[0]

    print_table_row(outfile, title, line)

    for i in range(1, len(header_names[0])):
        line = ''
        for name in header_names:
            line += ' ' + name[i]

        print_table_row(outfile, '', line)

    print_table_divider(outfile, num_cols)


def print_table_body(outfile, num_cols, ini_files, ini_data, default_features):
    """ Print out the body of the table. Each row is a NIC feature. """

    for feature, _ in default_features:
        line = ''

        for ini_filename in ini_files:
            line += ' ' + ini_data[ini_filename][feature]

        print_table_row(outfile, feature, line)

    print_table_divider(outfile, num_cols)


def print_table_row(outfile, feature, line):
    """ Print a single row of the table with fixed formatting. """
    line = line.rstrip()
    print('   {:<{}}{}'.format(feature, feature_str_len, line), file=outfile)


def print_table_divider(outfile, num_cols):
    """ Print the table divider line. """
    line = ' '
    column_dividers = ['='] * num_cols
    line += ' '.join(column_dividers)

    feature = '=' * feature_str_len

    print_table_row(outfile, feature, line)


def print_table_css(outfile, table_id):
    template = """
.. raw:: html

   <style>
      .wy-nav-content {
         opacity: .99;
      }
      table#idx {
         cursor: default;
         overflow: hidden;
      }
      table#idx p {
         margin: 0;
         line-height: inherit;
      }
      table#idx th, table#idx td {
         text-align: center;
         border: solid 1px #ddd;
      }
      table#idx th {
         padding: 0.5em 0;
      }
      table#idx th, table#idx th p {
         font-size: 11px;
         white-space: pre-wrap;
         vertical-align: top;
         min-width: 0.9em;
      }
      table#idx col:first-child {
         width: 0;
      }
      table#idx th:first-child {
         vertical-align: bottom;
      }
      table#idx td {
         padding: 1px;
      }
      table#idx td, table#idx td p {
         font-size: 11px;
      }
      table#idx td:first-child {
         padding-left: 1em;
         text-align: left;
      }
      table#idx tr:nth-child(2n-1) td {
         background-color: rgba(210, 210, 210, 0.2);
      }
      table#idx th:not(:first-child):hover,
      table#idx td:not(:first-child):hover {
         position: relative;
      }
      table#idx th:not(:first-child):hover::after,
      table#idx td:not(:first-child):hover::after {
         content: '';
         height: 6000px;
         top: -3000px;
         width: 100%;
         left: 0;
         position: absolute;
         z-index: -1;
         background-color: #ffb;
      }
      table#idx tr:hover td {
         background-color: #ffb;
      }
   </style>
"""
    print(template.replace("idx", "id%d" % (table_id)), file=outfile)


def setup(app):
    table_file = dirname(__file__) + '/nics/overview_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in networking drivers',
                            'Feature')
    table_file = dirname(__file__) + '/nics/rte_flow_items_table.txt'
    generate_overview_table(table_file, 2,
                            'rte_flow items',
                            'rte_flow items availability in networking drivers',
                            'Item')
    table_file = dirname(__file__) + '/nics/rte_flow_actions_table.txt'
    generate_overview_table(table_file, 3,
                            'rte_flow actions',
                            'rte_flow actions availability in networking drivers',
                            'Action')
    table_file = dirname(__file__) + '/cryptodevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in crypto drivers',
                            'Feature')
    table_file = dirname(__file__) + '/cryptodevs/overview_cipher_table.txt'
    generate_overview_table(table_file, 2,
                            'Cipher',
                            'Cipher algorithms in crypto drivers',
                            'Cipher algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_auth_table.txt'
    generate_overview_table(table_file, 3,
                            'Auth',
                            'Authentication algorithms in crypto drivers',
                            'Authentication algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_aead_table.txt'
    generate_overview_table(table_file, 4,
                            'AEAD',
                            'AEAD algorithms in crypto drivers',
                            'AEAD algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_asym_table.txt'
    generate_overview_table(table_file, 5,
                            'Asymmetric',
                            'Asymmetric algorithms in crypto drivers',
                            'Asymmetric algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_os_table.txt'
    generate_overview_table(table_file, 6,
                            'OS',
                            'Operating systems support for crypto drivers',
                            'Operating system')
    table_file = dirname(__file__) + '/compressdevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in compression drivers',
                            'Feature')
    table_file = dirname(__file__) + '/regexdevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in regex drivers',
                            'Feature')
    table_file = dirname(__file__) + '/vdpadevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in vDPA drivers',
                            'Feature')
    table_file = dirname(__file__) + '/bbdevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in bbdev drivers',
                            'Feature')
    table_file = dirname(__file__) + '/gpus/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in GPU drivers',
                            'Feature')
    table_file = dirname(__file__) + '/eventdevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Scheduling Features',
                            'Features availability in eventdev drivers',
                            'Feature')
    table_file = dirname(__file__) + '/eventdevs/overview_rx_adptr_feature_table.txt'
    generate_overview_table(table_file, 2,
                            'Eth Rx adapter Features',
                            'Features availability for Ethdev Rx adapters',
                            'Feature')
    table_file = dirname(__file__) + '/eventdevs/overview_tx_adptr_feature_table.txt'
    generate_overview_table(table_file, 3,
                            'Eth Tx adapter Features',
                            'Features availability for Ethdev Tx adapters',
                            'Feature')
    table_file = dirname(__file__) + '/eventdevs/overview_crypto_adptr_feature_table.txt'
    generate_overview_table(table_file, 4,
                            'Crypto adapter Features',
                            'Features availability for Crypto adapters',
                            'Feature')
    table_file = dirname(__file__) + '/eventdevs/overview_timer_adptr_feature_table.txt'
    generate_overview_table(table_file, 5,
                            'Timer adapter Features',
                            'Features availability for Timer adapters',
                            'Feature')

    if Version(sphinx_version) < Version('1.3.1'):
        print('Upgrade sphinx to version >= 1.3.1 for '
              'improved Figure/Table number handling.',
              file=stderr)
        # Add a role to handle :numref: references.
        app.add_role('numref', numref_role)
        # Process the numref references once the doctree has been created.
        app.connect('doctree-resolved', process_numref)

    try:
        # New function in sphinx 1.8
        app.add_css_file('css/custom.css')
    except:
        app.add_stylesheet('css/custom.css')



#! /usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2020 Intel Corporation

import sys
import os
from glob import iglob

if len(sys.argv) != 2:
    print("Usage: {0} <path-glob>[,<path-glob>[,...]]".format(sys.argv[0]))
    sys.exit(1)

root = os.path.join(os.getenv('MESON_SOURCE_ROOT', '.'),
                    os.getenv('MESON_SUBDIR', '.'))

for path in sys.argv[1].split(','):
    if path:
        for p in iglob(os.path.join(root, path)):
            if os.path.isdir(p):
                print(os.path.relpath(p, start=root).replace('\\', '/'))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2021 Intel Corporation
# Copyright(c) 2022-2023 PANTHEON.tech s.r.o.
# Copyright(c) 2022 University of New Hampshire
# Copyright(c) 2024 Arm Limited

"""Environment variables and command line arguments parsing.

This is a simple module utilizing the built-in argparse module to parse command line arguments,
augment them with values from environment variables and make them available across the framework.

The command line value takes precedence, followed by the environment variable value,
followed by the default value defined in this module.

The command line arguments along with the supported environment variables are:

.. option:: --config-file
.. envvar:: DTS_CFG_FILE

    The path to the YAML test run configuration file.

.. option:: --output-dir, --output
.. envvar:: DTS_OUTPUT_DIR

    The directory where DTS logs and results are saved.

.. option:: --compile-timeout
.. envvar:: DTS_COMPILE_TIMEOUT

    The timeout for compiling DPDK.

.. option:: -t, --timeout
.. envvar:: DTS_TIMEOUT

    The timeout for all DTS operation except for compiling DPDK.

.. option:: -v, --verbose
.. envvar:: DTS_VERBOSE

    Set to any value to enable logging everything to the console.

.. option:: --dpdk-tree
.. envvar:: DTS_DPDK_TREE

    The path to the DPDK source tree directory to test. Cannot be used in conjunction with
    --tarball.

.. option:: --tarball, --snapshot
.. envvar:: DTS_DPDK_TARBALL

    The path to the DPDK source tarball to test. DPDK must be contained in a folder with the same
    name as the tarball file. Cannot be used in conjunction with --dpdk-tree.

.. option:: --remote-source
.. envvar:: DTS_REMOTE_SOURCE

    Set this option if either the DPDK source tree or tarball to be used are located on the SUT
    node. Can only be used with --dpdk-tree or --tarball.

.. option:: --precompiled-build-dir
.. envvar:: DTS_PRECOMPILED_BUILD_DIR

    Define the subdirectory under the DPDK tree root directory or tarball where the pre-compiled
    binaries are located.

.. option:: --test-suite
.. envvar:: DTS_TEST_SUITES

        A test suite with test cases which may be specified multiple times.
        In the environment variable, the suites are joined with a comma.

.. option:: --re-run, --re_run
.. envvar:: DTS_RERUN

    Re-run each test case this many times in case of a failure.

.. option:: --random-seed
.. envvar:: DTS_RANDOM_SEED

    The seed to use with the pseudo-random generator. If not specified, the configuration value is
    used instead. If that's also not specified, a random seed is generated.

The module provides one key module-level variable:

Attributes:
    SETTINGS: The module level variable storing framework-wide DTS settings.

Typical usage example::

  from framework.settings import SETTINGS
  foo = SETTINGS.foo
"""

import argparse
import os
import sys
from argparse import Action, ArgumentDefaultsHelpFormatter, _get_action_name
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable

from pydantic import ValidationError

from .config import (
    DPDKLocation,
    LocalDPDKTarballLocation,
    LocalDPDKTreeLocation,
    RemoteDPDKTarballLocation,
    RemoteDPDKTreeLocation,
    TestSuiteConfig,
)


@dataclass(slots=True)
class Settings:
    """Default framework-wide user settings.

    The defaults may be modified at the start of the run.
    """

    #:
    config_file_path: Path = Path(__file__).parent.parent.joinpath("conf.yaml")
    #:
    output_dir: str = "output"
    #:
    timeout: float = 15
    #:
    verbose: bool = False
    #:
    dpdk_location: DPDKLocation | None = None
    #:
    precompiled_build_dir: str | None = None
    #:
    compile_timeout: float = 1200
    #:
    test_suites: list[TestSuiteConfig] = field(default_factory=list)
    #:
    re_run: int = 0
    #:
    random_seed: int | None = None


SETTINGS: Settings = Settings()


#: Attribute name representing the env variable name to augment :class:`~argparse.Action` with.
_ENV_VAR_NAME_ATTR = "env_var_name"
#: Attribute name representing the value origin to augment :class:`~argparse.Action` with.
_IS_FROM_ENV_ATTR = "is_from_env"

#: The prefix to be added to all of the environment variables.
_ENV_PREFIX = "DTS_"


def _make_env_var_name(action: Action, env_var_name: str | None) -> str:
    """Make and assign an environment variable name to the given action."""
    env_var_name = f"{_ENV_PREFIX}{env_var_name or action.dest.upper()}"
    setattr(action, _ENV_VAR_NAME_ATTR, env_var_name)
    return env_var_name


def _get_env_var_name(action: Action) -> str | None:
    """Get the environment variable name of the given action."""
    return getattr(action, _ENV_VAR_NAME_ATTR, None)


def _set_is_from_env(action: Action) -> None:
    """Make the environment the given action's value origin."""
    setattr(action, _IS_FROM_ENV_ATTR, True)


def _is_from_env(action: Action) -> bool:
    """Check if the given action's value originated from the environment."""
    return getattr(action, _IS_FROM_ENV_ATTR, False)


def _is_action_in_args(action: Action) -> bool:
    """Check if the action is invoked in the command line arguments."""
    for option in action.option_strings:
        if option in sys.argv:
            return True
    return False


def _add_env_var_to_action(
    action: Action,
    env_var_name: str | None = None,
) -> None:
    """Add an argument with an environment variable to the parser."""
    env_var_name = _make_env_var_name(action, env_var_name)

    if not _is_action_in_args(action):
        env_var_value = os.environ.get(env_var_name)
        if env_var_value is not None:
            _set_is_from_env(action)
            sys.argv[1:0] = [action.format_usage(), env_var_value]


class _DTSArgumentParser(argparse.ArgumentParser):
    """ArgumentParser with a custom error message.

    This custom version of ArgumentParser changes the error message to accurately reflect the origin
    of the value of its arguments. If it was supplied through the command line nothing changes, but
    if it was supplied as an environment variable this is correctly communicated.
    """

    def find_action(
        self, action_dest: str, filter_fn: Callable[[Action], bool] | None = None
    ) -> Action | None:
        """Find and return an action by its destination variable name.

        Arguments:
            action_dest: the destination variable name of the action to find.
            filter_fn: if an action is found it is passed to this filter function, which must
                return a boolean value.
        """
        it = (action for action in self._actions if action.dest == action_dest)
        action = next(it, None)

        if action and filter_fn:
            return action if filter_fn(action) else None

        return action

    def error(self, message):
        """Augments :meth:`~argparse.ArgumentParser.error` with environment variable awareness."""
        for action in self._actions:
            if _is_from_env(action):
                action_name = _get_action_name(action)
                env_var_name = _get_env_var_name(action)
                env_var_value = os.environ.get(env_var_name)

                message = message.replace(
                    f"argument {action_name}",
                    f"environment variable {env_var_name} (value: {env_var_value})",
                )

        print(f"{self.prog}: error: {message}\n", file=sys.stderr)
        self.exit(2, "For help and usage, " "run the command with the --help flag.\n")


class _EnvVarHelpFormatter(ArgumentDefaultsHelpFormatter):
    """Custom formatter to add environment variables to the help page."""

    def _get_help_string(self, action):
        """Overrides :meth:`ArgumentDefaultsHelpFormatter._get_help_string`."""
        help = super()._get_help_string(action)

        env_var_name = _get_env_var_name(action)
        if env_var_name is not None:
            help = f"[{env_var_name}] {help}"

            env_var_value = os.environ.get(env_var_name)
            if env_var_value is not None:
                help = f"{help} (env value: {env_var_value})"

        return help


def _required_with_one_of(parser: _DTSArgumentParser, action: Action, *required_dests: str) -> None:
    """Verify that `action` is listed together with at least one of `required_dests`.

    Verify that when `action` is among the command-line arguments or
    environment variables, at least one of `required_dests` is also among
    the command-line arguments or environment variables.

    Args:
        parser: The custom ArgumentParser object which contains `action`.
        action: The action to be verified.
        *required_dests: Destination variable names of the required arguments.

    Raises:
        argparse.ArgumentTypeError: When none of the required_dest are defined.

    Example:
        We have ``--option1`` and we only want it to be a passed alongside
        either ``--option2`` or ``--option3`` (meaning if ``--option1`` is
        passed without either ``--option2`` or ``--option3``, that's an error).

        parser = _DTSArgumentParser()
        option1_arg = parser.add_argument('--option1', dest='option1', action='store_true')
        option2_arg = parser.add_argument('--option2', dest='option2', action='store_true')
        option2_arg = parser.add_argument('--option3', dest='option3', action='store_true')

        _required_with_one_of(parser, option1_arg, 'option2', 'option3')
    """
    if _is_action_in_args(action):
        for required_dest in required_dests:
            required_action = parser.find_action(required_dest)
            if required_action is None:
                continue

            if _is_action_in_args(required_action):
                return None

        raise argparse.ArgumentTypeError(
            f"The '{action.dest}' is required at least with one of '{', '.join(required_dests)}'."
        )


def _get_parser() -> _DTSArgumentParser:
    """Create the argument parser for DTS.

    Command line options take precedence over environment variables, which in turn take precedence
    over default values.

    Returns:
        _DTSArgumentParser: The configured argument parser with defined options.
    """
    parser = _DTSArgumentParser(
        description="Run DPDK test suites. All options may be specified with the environment "
        "variables provided in brackets. Command line arguments have higher priority.",
        formatter_class=_EnvVarHelpFormatter,
        allow_abbrev=False,
    )

    action = parser.add_argument(
        "--config-file",
        default=SETTINGS.config_file_path,
        type=Path,
        help="The configuration file that describes the test cases, SUTs and DPDK build configs.",
        metavar="FILE_PATH",
        dest="config_file_path",
    )
    _add_env_var_to_action(action, "CFG_FILE")

    action = parser.add_argument(
        "--output-dir",
        "--output",
        default=SETTINGS.output_dir,
        help="Output directory where DTS logs and results are saved.",
        metavar="DIR_PATH",
    )
    _add_env_var_to_action(action)

    action = parser.add_argument(
        "-t",
        "--timeout",
        default=SETTINGS.timeout,
        type=float,
        help="The default timeout for all DTS operations except for compiling DPDK.",
        metavar="SECONDS",
    )
    _add_env_var_to_action(action)

    action = parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=SETTINGS.verbose,
        help="Specify to enable verbose output, logging all messages to the console.",
    )
    _add_env_var_to_action(action)

    dpdk_build = parser.add_argument_group(
        "DPDK Build Options",
        description="Arguments in this group (and subgroup) will be applied to a "
        "DPDKLocation when the DPDK tree, tarball or revision will be provided, "
        "other arguments like remote source and build dir are optional. A DPDKLocation "
        "from settings are used instead of from config if construct successful.",
    )

    dpdk_source = dpdk_build.add_mutually_exclusive_group()
    action = dpdk_source.add_argument(
        "--dpdk-tree",
        help="The path to the DPDK source tree directory to test. Cannot be used in conjunction "
        "with --tarball.",
        metavar="DIR_PATH",
        dest="dpdk_tree_path",
    )
    _add_env_var_to_action(action, "DPDK_TREE")

    action = dpdk_source.add_argument(
        "--tarball",
        "--snapshot",
        help="The path to the DPDK source tarball to test. DPDK must be contained in a folder with "
        "the same name as the tarball file. Cannot be used in conjunction with --dpdk-tree.",
        metavar="FILE_PATH",
        dest="dpdk_tarball_path",
    )
    _add_env_var_to_action(action, "DPDK_TARBALL")

    action = dpdk_build.add_argument(
        "--remote-source",
        action="store_true",
        default=False,
        help="Set this option if either the DPDK source tree or tarball to be used are located on "
        "the SUT node. Can only be used with --dpdk-tree or --tarball.",
    )
    _add_env_var_to_action(action)
    _required_with_one_of(parser, action, "dpdk_tarball_path", "dpdk_tree_path")

    action = dpdk_build.add_argument(
        "--precompiled-build-dir",
        help="Define the subdirectory under the DPDK tree root directory or tarball where the "
        "pre-compiled binaries are located.",
        metavar="DIR_NAME",
    )
    _add_env_var_to_action(action)

    action = parser.add_argument(
        "--compile-timeout",
        default=SETTINGS.compile_timeout,
        type=float,
        help="The timeout for compiling DPDK.",
        metavar="SECONDS",
    )
    _add_env_var_to_action(action)

    action = parser.add_argument(
        "--test-suite",
        action="append",
        nargs="+",
        metavar=("TEST_SUITE", "TEST_CASES"),
        default=SETTINGS.test_suites,
        help="A list containing a test suite with test cases. "
        "The first parameter is the test suite name, and the rest are test case names, "
        "which are optional. May be specified multiple times. To specify multiple test suites in "
        "the environment variable, join the lists with a comma. "
        "Examples: "
        "--test-suite suite case case --test-suite suite case ... | "
        "DTS_TEST_SUITES='suite case case, suite case, ...' | "
        "--test-suite suite --test-suite suite case ... | "
        "DTS_TEST_SUITES='suite, suite case, ...'",
        dest="test_suites",
    )
    _add_env_var_to_action(action)

    action = parser.add_argument(
        "--re-run",
        "--re_run",
        default=SETTINGS.re_run,
        type=int,
        help="Re-run each test case the specified number of times if a test failure occurs.",
        metavar="N_TIMES",
    )
    _add_env_var_to_action(action, "RERUN")

    action = parser.add_argument(
        "--random-seed",
        type=int,
        help="The seed to use with the pseudo-random generator. If not specified, the configuration"
        " value is used instead. If that's also not specified, a random seed is generated.",
        metavar="NUMBER",
    )
    _add_env_var_to_action(action)

    return parser


def _process_dpdk_location(
    parser: _DTSArgumentParser,
    dpdk_tree: str | None,
    tarball: str | None,
    remote: bool,
) -> DPDKLocation | None:
    """Process and validate DPDK build arguments.

    Ensures that either `dpdk_tree` or `tarball` is provided. Validate existence and format of
    `dpdk_tree` or `tarball` on local filesystem, if `remote` is False. Constructs and returns
    any valid :class:`DPDKLocation` with the provided parameters if validation is successful.

    Args:
        dpdk_tree: The path to the DPDK source tree directory.
        tarball: The path to the DPDK tarball.
        remote: If :data:`True`, `dpdk_tree` or `tarball` is located on the SUT node, instead of the
            execution host.

    Returns:
        A DPDK location if construction is successful, otherwise None.
    """
    if dpdk_tree:
        action = parser.find_action("dpdk_tree", _is_from_env)

        try:
            if remote:
                return RemoteDPDKTreeLocation.model_validate({"dpdk_tree": dpdk_tree})
            else:
                return LocalDPDKTreeLocation.model_validate({"dpdk_tree": dpdk_tree})
        except ValidationError as e:
            print(
                "An error has occurred while validating the DPDK tree supplied in the "
                f"{'environment variable' if action else 'arguments'}:",
                file=sys.stderr,
            )
            print(e, file=sys.stderr)
            sys.exit(1)

    if tarball:
        action = parser.find_action("tarball", _is_from_env)

        try:
            if remote:
                return RemoteDPDKTarballLocation.model_validate({"tarball": tarball})
            else:
                return LocalDPDKTarballLocation.model_validate({"tarball": tarball})
        except ValidationError as e:
            print(
                "An error has occurred while validating the DPDK tarball supplied in the "
                f"{'environment variable' if action else 'arguments'}:",
                file=sys.stderr,
            )
            print(e, file=sys.stderr)
            sys.exit(1)

    return None


def _process_test_suites(
    parser: _DTSArgumentParser, args: list[list[str]]
) -> list[TestSuiteConfig]:
    """Process the given argument to a list of :class:`TestSuiteConfig` to execute.

    Args:
        args: The arguments to process. The args is a string from an environment variable
              or a list of from the user input containing tests suites with tests cases,
              each of which is a list of [test_suite, test_case, test_case, ...].

    Returns:
        A list of test suite configurations to execute.
    """
    action = parser.find_action("test_suites", _is_from_env)
    if action:
        # Environment variable in the form of "SUITE1 CASE1 CASE2, SUITE2 CASE1, SUITE3, ..."
        args = [suite_with_cases.split() for suite_with_cases in args[0][0].split(",")]

    try:
        return [
            TestSuiteConfig(test_suite=test_suite, test_cases=test_cases)
            for [test_suite, *test_cases] in args
        ]
    except ValidationError as e:
        print(
            "An error has occurred while validating the test suites supplied in the "
            f"{'environment variable' if action else 'arguments'}:",
            file=sys.stderr,
        )
        print(e, file=sys.stderr)
        sys.exit(1)


def get_settings() -> Settings:
    """Create new settings with inputs from the user.

    The inputs are taken from the command line and from environment variables.

    Returns:
        The new settings object.
    """
    parser = _get_parser()

    args = parser.parse_args()

    args.dpdk_location = _process_dpdk_location(
        parser, args.dpdk_tree_path, args.dpdk_tarball_path, args.remote_source
    )
    args.test_suites = _process_test_suites(parser, args.test_suites)

    kwargs = {k: v for k, v in vars(args).items() if hasattr(SETTINGS, k)}
    return Settings(**kwargs)
